<!doctype html>
<html lang="en">
<head>
  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <title>{% block title %}tit{% endblock %}</title>
  <link rel="stylesheet" href="/assets/style.css">
</head>
<body>
  <a class="skip-link" href="#main">Skip to content</a>
  <header class="site-header">
    <a class="brand" href="/" aria-label="tit home">tit</a>
    <nav aria-label="Primary">
      <a href="/">Home</a>
    </nav>
  </header>
  <main id="main">
{% block content %}{% endblock %}
  </main>
  <footer>
    <small>Request <code>{{ request_id }}</code></small>
  </footer>
</body>
</html>
